#!/usr/bin/tclsh
#
# Authors:
# 2008, SHY <abbagom@gmail.com>
# Authentication, File transfer over TLS, Revision Control.

# Set System encoding to use Korean well.
encoding system utf-8

# Read system required packages
package require tls
package require md5crypt
package require base64
package require aes
package require sum
package require uri::urn

# Read client configuration
# This configuration file parser is from Mike Tuxford on comp.lang.tcl in Dec. 2002. See the http://wiki.tcl.tk/2438 for mores.
proc parseConfig {configfile} {
    if {[file exists $configfile]} {
	set fd [open $configfile r]
    } else {
	puts "Error : Failed to read configuration file. Check the file is exist or its permission."
	exit
    }
    while {![eof $fd]} {
	set data [gets $fd]
	# Skip the commented or null line
	if {[string index $data 0] == "#" || \
		[string index $data 0] == " " || \
		[string index $data 0] == ""} {
	    continue
	}
	# Set the "Config" array, using configuration file contents
	switch [lindex $data 0] {
	    DefaultServerIP {
		global Config
		set Config(DefaultServerIP) [lindex $data 1]
	    }
	    DefaultServerPort {
		global Config
		set Config(DefaultServerPort) [lindex $data 1]
	    }
	    DefaultAccount {
		global Config
		set Config(DefaultAccount) [lrange $data 1 end]
	    }

	    default {
		global Config
		set Config([lindex $data 0]) [lrange $data 1 end]
	    }
	}
    }
    catch {close $fd}
    return
}
proc saveConfig {configfile item value} {
    # Read whole config file contents
    if {[file exists $configfile]} {
	set fd [open $configfile r]
	set config_data [read $fd]
	close $fd
    } else {
	return "Error : Failed to read/write configuration file. Check the existence of file and its permission."
    }

    # Change the config line, by assigned item/value.
    set config_data [regsub -all -lineanchor "^$item .*?$" $config_data "$item $value"]

    # Write changed configuration to file.
    file rename -force $configfile [format "%s.bak" $configfile]
    set fd [open $configfile w]
    puts -nonewline $fd $config_data
    close $fd

    # End
    return "INFO : Configuration is changed succefully."
}
parseConfig ShareBear_client.cfg

proc connect {server_ip server_port} {
    # When ip/port is not assigned, use the default one.
    if {$server_ip == ""} {
	set server_ip $::Config(DefaultServerIP)
    }
    if {$server_port == ""} {
	set server_port $::Config(DefaultServerPort)
    }

    # Save ip/port information to default.
    set ::Config(DefaultServerIP) $server_ip
    set ::Config(DefaultServerPort) $server_port
    saveConfig ShareBear_client.cfg DefaultServerIP $server_ip
    saveConfig ShareBear_client.cfg DefaultServerPort $server_port

    # Check connection status
    if {[connected_or_not] == "Connected"} {
	return "Fail : Server is already connected."
    }

    # Connect & Configure that.
    if {[catch {set ::s [::tls::socket $server_ip $server_port]}]} {
	return "Fail : Fail to connect to $server_ip : $server_port"
    }
    fconfigure $::s -blocking 0 -buffering line -translation binary -encoding utf-8

    # Get wellcome message from server
    for {gets $::s line} {[string equal $line ""]} {gets $::s line} {after 100} ;# Wait until some data is arrived on channel.
    append Rx_buffer $line
    while {[gets $::s line] >= 0} {
	append Rx_buffer "\n"
	append Rx_buffer $line
    }
    return $Rx_buffer
}

proc disconnect {} {
    # Check connection status
    if {[connected_or_not] == "NotConnected"} {
	return "Fail : Not Connected."
    }

    # Close the connection
    puts $::s "bye"

    # Read server response
    for {gets $::s line} {[string equal $line ""]} {gets $::s line} {after 100}
    append Rx_buffer $line
    while {[gets $::s line] >= 0} {
	append Rx_buffer "\n"
	append Rx_buffer $line
    }

    return $Rx_buffer
}

proc connected_or_not {} {
    if {[catch {tls::handshake $::s} message_connectionstatus]} {
	return "NotConnected" ;# There's some error on tls::handshake command or ::s var.
    } elseif {$message_connectionstatus == 0} {
	return "NotConnected" ;# tls::handshake command returned 0 when connection is not established.
    } elseif {$message_connectionstatus == 1} { ;#succeed for tls handshake
	if {[eof $::s]}	{
	    return "NotConnected" ;#but socket is closed at server side.
	} else {
	    return "Connected" ;#socket is open
	}
    }
}

proc command {arg} {
    # Check connection status
    if {[connected_or_not] == "NotConnected"} {
	return "Fail : Not Connected."
    }

    while {[gets $::s line] >= 0} { puts $line } ;# Clear the channel buffer
    puts $::s $arg ;# Send command

    # Read server response
    for {gets $::s line} {[string equal $line ""]} {gets $::s line} {after 100}
    append Rx_buffer $line
    while {[gets $::s line] >= 0} {
	append Rx_buffer "\n"
	append Rx_buffer $line
    }

    return $Rx_buffer
}

proc I_need_salt {} {
    while {[gets $::s line] >= 0} { puts $line }
    puts $::s share_the_salt
    for {gets $::s line} {[string equal $line ""]} {gets $::s line} {after 100}
    set ::salt $line
    return "Salt from server : $::salt"
}

proc register_account {id password email etc} {
    # Check connection status
    if {[connected_or_not] == "NotConnected"} {
	return "Fail : Not Connected."
    }

    I_need_salt ;# read salt to communicate password

    # When email or etc is empty, replace that automatically.
    if {$email == ""} {
	set email _null_
    }
    if {$etc == ""} {
	set etc _null_
    }

    # Make a id/email/etc item in url-encrypted format to avoid space and anything problem.
    set id [uri::urn::quote [encoding convertto utf-8 $id]]
    set email [uri::urn::quote [encoding convertto utf-8 $email]]
    set etc [uri::urn::quote [encoding convertto utf-8 $etc]]

    #Hashed Password Generation : md5crypt by user password and user ID.
    set hashed_password [md5crypt::md5crypt $password $id]
    #Hashed Password Encryption : aes by salt as password
    set key $::salt
    set encrypted_hashed_password [base64::encode -maxlen 0 [aes::aes -dir encrypt -key $key -mode ecb $hashed_password]]
    #Token Generation : md5crypt by hashed password and user ID
    set token [md5crypt::md5crypt $hashed_password $id]
    #Send user name, Encrypted hashed password, Token and mores and wait for response
    while {[gets $::s line] >= 0} { puts $line }
    puts $::s "register_account $id $encrypted_hashed_password $token $email $etc"

    # Read server response
    for {gets $::s line} {[string equal $line ""]} {gets $::s line} {after 100}
    append Rx_buffer $line
    while {[gets $::s line] >= 0} {
	append Rx_buffer "\n"
	append Rx_buffer $line
    }

    return $Rx_buffer
}

proc authentication {id password} {
    # Check connection status
    if {[connected_or_not] == "NotConnected"} {
	return "Fail : Not Connected."
    }

    # When ID/Password is not defined, use default one.
    if {$id == ""} {
	set id $::Config(DefaultAccount)
	set password ""
    }

    # Save user ID as default, to use it later.
    set ::Config(DefaultAccount) $id
    saveConfig ShareBear_client.cfg DefaultAccount $id

    # Make a id in url-encrypted format to avoid space and anything problem.
    set id [uri::urn::quote [encoding convertto utf-8 $id]]

    I_need_salt ;# read salt to communicate password

    #Hashed Password Generation : md5crypt by user password and user ID.
    set hashed_password [md5crypt::md5crypt $password $id]
    #Hashed Password Encryption : aes by salt as password
    set key $::salt
    set encrypted_hashed_password [base64::encode -maxlen 0 [aes::aes -dir encrypt -key $key -mode ecb $hashed_password]]
    #Token Generation : md5crypt by hashed password and user ID
    set token [md5crypt::md5crypt $hashed_password $id]

    #Send user name, Encrypted hashed password, Token and mores and wait for response
    while {[gets $::s line] >= 0} { puts $line }
    puts $::s "authentication $id $encrypted_hashed_password $token"

    # Read server response
    for {gets $::s line} {[string equal $line ""]} {gets $::s line} {after 100}
    append Rx_buffer $line
    while {[gets $::s line] >= 0} {
	append Rx_buffer "\n"
	append Rx_buffer $line
    }

    return $Rx_buffer
}

proc I_need_chunksize {} {
    while {[gets $::s line] >= 0} { puts $line }
    puts $::s share_the_chunksize
    for {gets $::s line} {[string equal $line ""]} {gets $::s line} {after 100}
    set ::chunksize $line
    return "Chunksize from server : $::chunksize"
}

proc upload {NewORFileID FullyQualifiedFileList TagList Description} { 
    # Check connection status
    if {[connected_or_not] == "NotConnected"} {
	return "Fail : Not Connected."
    }

    # Check whether file is exist or not
    if {[llength $FullyQualifiedFileList] == 0} {
	return "Error : Filelist is empty."
    }
    foreach FullyQualifiedFileName $FullyQualifiedFileList {
	if {![file exist $FullyQualifiedFileName]} {
	    return "Error : File(s) is not exist."
	}
    }

    I_need_chunksize

    # FileList information to server has items which has unique file name. 
    # File path of each file, should be also removed from FileList.
    # FileName(which is element of FileList) should be url-encoded to avoid white space problem of list.
    set FileList_encoded {}
    foreach FullyQualifiedFileName $FullyQualifiedFileList {
	lappend FileList_encoded [uri::urn::quote [encoding convertto utf-8 [file tail $FullyQualifiedFileName]]]
    }

    # File Description should be url-encoded to be recognized as one element
    set Description [uri::urn::quote [encoding convertto utf-8 $Description]]

    # Empty socket, then send command to server.
    while {[gets $::s Rx_buffer] >= 0} { puts $Rx_buffer }
    puts $::s "file_upload {$NewORFileID} {$FileList_encoded} {$TagList} {$Description}" ;# List should be encapsulated by parenthesis.
    for {gets $::s Rx_buffer} {[string equal $Rx_buffer ""]} {gets $::s Rx_buffer} {after 100}

    # Response for server request, until file transfer is completed.
    while {$Rx_buffer != "Info : File transfer is completed." } {
	if {[lindex $Rx_buffer 0] == "QUERY_FileInfo"} { ;# QUERY_FileInfo <file name> FileSize FileChecksum
	    set FullyQualifiedFileName [lindex $FullyQualifiedFileList [lsearch $FileList_encoded [lindex $Rx_buffer 1]]]
	    set FileSize [file size $FullyQualifiedFileName]
	    set FileChecksum [crc::sum -file $FullyQualifiedFileName]
	    puts $::s "FileInfo [lindex $Rx_buffer 1] $FileSize $FileChecksum" ;# FileInfo <file name> <file size> <file checksum>

	    puts "UPLOAD \([lindex $Rx_buffer 1]\): Upload is started." ;# Display status

	} elseif {[lindex $Rx_buffer 0] == "QUERY_ChunkInfo"} { ;# QUERY_ChunkInfo <file name> <chunk number> ChunkCheckSum
	    set FullyQualifiedFileName [lindex $FullyQualifiedFileList [lsearch $FileList_encoded [lindex $Rx_buffer 1]]]
	    set ChunkNumber [lindex $Rx_buffer 2]

	    if {[GetChunk $FullyQualifiedFileName $::chunksize $ChunkNumber] == "Info : Completed to make chunk file."} {
		set ChunkChecksum [crc::sum -file ShareBear.chk]
	    }

	    puts $::s "ChunkInfo [lindex $Rx_buffer 1] [lindex $Rx_buffer 2] $ChunkChecksum" ;# ChunkInfo <file name> <chunk number> <chunk checksum>

	    puts "UPLOAD \([lindex $Rx_buffer 1]\): $ChunkNumber / [expr {int(ceil($FileSize/[format %f $::chunksize]))}] in progress." ;# Display status

	} elseif {[lindex $Rx_buffer 0] == "QUERY_ChunkUpload"} { ;# QUERY_ChunkUpload <file name> <chunk number>
	    set FullyQualifiedFileName [lindex $FullyQualifiedFileList [lsearch $FileList_encoded [lindex $Rx_buffer 1]]]
	    set ChunkNumber [lindex $Rx_buffer 2]

	    if {[GetChunk $FullyQualifiedFileName $::chunksize $ChunkNumber] == "Info : Completed to make chunk file."} {
		set Chunk [open ShareBear.chk r]
		fconfigure $Chunk -translation binary
		fconfigure $::s -buffering line -blocking 0 -translation binary -encoding binary
		fcopy $Chunk $::s
		close $Chunk
		fconfigure $::s -buffering line -blocking 0 -translation binary -encoding utf-8
	    }

	    puts "UPLOAD \([lindex $Rx_buffer 1]\): $ChunkNumber / [expr {int(ceil($FileSize/[format %f $::chunksize]))}] is uploaded." ;# Display status

	}

	# Read Command from Server, for next process
	for {gets $::s Rx_buffer} {[string equal $Rx_buffer ""]} {gets $::s Rx_buffer} {after 100}
    }

    # Delete Chunk File
    file delete ShareBear.chk

    # Print Server Message
    return $Rx_buffer
}


proc GetChunk {FileName ChunkSize ChunkNumber} {
    # Open Filestream and configure that.
    set fp [open $FileName r] ;# Fully qualified filename
    set Chunk [open ShareBear.chk w]

    fconfigure $fp -encoding binary -translation binary
    fconfigure $Chunk -encoding binary -translation binary
    
    seek $fp [expr (($ChunkNumber - 1) * $ChunkSize)] start

    # Calculate the total number of chunk
    set FileSize [file size $FileName]
    set TotalChunkNumber [expr {int(ceil($FileSize/[format %f $ChunkSize]))}]

    # Copy the Chunk from File
    if {$ChunkNumber < $TotalChunkNumber} {
	fcopy $fp $Chunk -size $ChunkSize
    } elseif {$ChunkNumber == $TotalChunkNumber} {
	fcopy $fp $Chunk -size [expr $FileSize % $ChunkSize ]
    } else {
	return "Error : invalid chunk number was given."
    }

    # Close Filestream
    close $fp
    close $Chunk

    return "Info : Completed to make chunk file."
}

proc Get_TagList {} { ;# Ask server to send existing taglist
    # Check connection status
    if {[connected_or_not] == "NotConnected"} {
	return "Fail : Not Connected."
    }

    # Get the taglist from server
    while {[gets $::s line] >= 0} { puts $line }
    puts $::s "send_taglist"
    for {gets $::s line} {[string equal $line ""]} {gets $::s line} {after 100}
    set tag_list_encoded $line

    # Decoding tag list
    foreach Tag $tag_list_encoded {
	lappend tag_list_decoded [encoding convertfrom utf-8 [uri::urn::unquote $Tag]]
    }

    # return decoded tag-list
    return $tag_list_decoded
}

proc Get_FileList {TagList Selection_AndOr} { ;# Request server to send FileID, FileRevision, FileList using TagList which I informed.
    # Check connection status
    if {[connected_or_not] == "NotConnected"} {
	return "Fail : Not Connected."
    }

    # Empty socket, then send command to server.
    while {[gets $::s line] >= 0} { puts $line }
    puts $::s "send_filelist {$TagList} $Selection_AndOr" ;# List should be encapsulated by parenthesis. Selection_AndOr : If 'and', server will return the file list which has all tag element of client specified. If 'or', server will return the file list which has one of tag of client specified.
    for {gets $::s line} {[string equal $line ""]} {gets $::s line} {after 100} ;# Wait until some data is arrived on channel.
    lappend Rx_buffer $line ;# Make a received line to list
    while {[gets $::s line] >= 0} { lappend Rx_buffer $line } ;# print whole arrived data to Rx_buffer

    # Split the data and decoding the filename
    if { [lindex [lindex $Rx_buffer 0] 0] != "Error" &&
	 [lindex [lindex $Rx_buffer 0] 0] != "ERROR" &&
	 [lindex [lindex $Rx_buffer 0] 0] != "Fail" &&
	 [lindex [lindex $Rx_buffer 0] 0] != "FAIL" &&
	 [lindex [lindex $Rx_buffer 0] 0] != "Info" &&
	 [lindex [lindex $Rx_buffer 0] 0] != "INFO" } {
	foreach line $Rx_buffer {
	    scan $line "FileID:%d FileRevision:%d FileList:%s" FileID FileRevision FileList
	    set FileList_decoded {}
	    foreach FileName [split $FileList ","] {
		lappend FileList_decoded [encoding convertfrom utf-8 [uri::urn::unquote $FileName]]
	    }
	    set FileList_decoded [join $FileList_decoded ","]
	    lappend FileID_FileRevision_FileList [format "FileID:%d FileRevision:%d FileList:%s" $FileID $FileRevision $FileList_decoded] ;# Make a list of formatted information : FileID-FileRevision-FileList
	}
	set FileID_FileRevision_FileList [join $FileID_FileRevision_FileList "\n"]
    } else {
	set FileID_FileRevision_FileList [lindex $Rx_buffer 0] ;# Error/Fail/Info message.
    }

    # Print it
    return $FileID_FileRevision_FileList
}

proc Get_FileNames {FileID FileRevision} { ;# Get the filelist of specified FileID, FileRevision
    # Check connection status
    if {[connected_or_not] == "NotConnected"} {
	return "Fail : Not Connected."
    }

    # Empty socket, then send command to server.
    while {[gets $::s line] >= 0} { puts $line }
    puts $::s "send_filename $FileID $FileRevision"
    for {gets $::s line} {[string equal $line ""]} {gets $::s line} {after 100} ;# Wait until some data is arrived on channel.

    # Decoding each filename to original character
    if { [lindex $line 0] != "Error" &&
	 [lindex $line 0] != "ERROR" &&
	 [lindex $line 0] != "Fail" &&
	 [lindex $line 0] != "FAIL" &&
	 [lindex $line 0] != "Info" &&
	 [lindex $line 0] != "INFO" } {
	set FileNameList_decoded {}
	foreach FileName $line {
	    lappend FileNameList_decoded [encoding convertfrom utf-8 [uri::urn::unquote $FileName]]
	}
	return $FileNameList_decoded
    } else {
	return $line
    }
}

proc Get_FileDescription {FileID FileRevision} { ;# Request server to send FileDescription
    # Check connection status
    if {[connected_or_not] == "NotConnected"} {
	return "Fail : Not Connected."
    }

    # Empty socket, then send command to server.
    while {[gets $::s line] >= 0} { puts $line }
    puts $::s "send_filedescription $FileID $FileRevision"
    for {gets $::s line} {[string equal $line ""]} {gets $::s line} {after 100} ;# Wait until some data is arrived on channel.

    # Decoding description to original character
    set FileDescription [encoding convertfrom utf-8 [uri::urn::unquote $line]]

    # Print it
    return $FileDescription
}

proc Get_FileInformation {FileID FileRevision} { ;# Request server to send overall file information
    # Check connection status
    if {[connected_or_not] == "NotConnected"} {
	return "Fail : Not Connected."
    }

    # Empty socket, then send command to server.
    while {[gets $::s line] >= 0} { puts $line }
    puts $::s "send_fileinformation $FileID $FileRevision"
    for {gets $::s line} {[string equal $line ""]} {gets $::s line} {after 100} ;# Wait until some data is arrived on channel.

    # Decoding description to original character
    set FileInformation [encoding convertfrom utf-8 [uri::urn::unquote $line]]

    # Print it
    return $FileInformation
}

proc download {FileID FileRevision FileName WhereToDownload} {
    # Check connection status
    if {[connected_or_not] == "NotConnected"} {
	return "Fail : Not Connected."
    }

    I_need_chunksize ;# check chunksize to download

    # Check whether download directory is valid or not.
    if {![file isdirectory $WhereToDownload]} {
	return "Error : Download Directory name is invalid."
    }

    # Make a pseudo file @ download directory
    set FullyQualifiedFileName [file join $WhereToDownload $FileName]
    set fp [open $FullyQualifiedFileName w]
    close $fp

    # Request server to send file checksum. If file is not invalid, close the proc.
    set FileName_encoded [uri::urn::quote [encoding convertto utf-8 $FileName]]
    puts $::s "file_download $FileID $FileRevision $FileName_encoded"
    for {gets $::s FileInfo} {[string equal $FileInfo ""]} {gets $::s FileInfo} {after 100}
    if {$FileInfo == "Error : Not existed file is requested to download."} {
	return $FileInfo
    } else {
	set FileSize [lindex $FileInfo 2]
	set FileChecksum [lindex $FileInfo 3]
    }

    # Until actual file's checksum is same with checksum value from server, download the chunk checksum/chunk.
    while { [crc::sum -file $FullyQualifiedFileName] != $FileChecksum } {
	# Open the file Channel
	set fp [open $FullyQualifiedFileName w]
	fconfigure $fp -translation binary

	# Calculate the total number of chunk, Current Chunk
	set TotalChunkNumber [expr {int(ceil($FileSize/[format %f $::chunksize]))}]
	set CurrentChunkNumber 1

	# In order, Receive & Check the chunk by checksum. If it isn't valid, request it once more.
	while {$CurrentChunkNumber <= $TotalChunkNumber} {
	    # Request & Receive the chunk Info
	    puts $::s "QUERY_ChunkInfo $FileName_encoded $CurrentChunkNumber ChunkCheckSum"
	    for {gets $::s ChunkInfo} {[string equal $ChunkInfo ""]  && (![eof $::s])} {gets $::s ChunkInfo} {after 100} ;# Wait until some data is arrived.
	    if {[llength $ChunkInfo] == 4
		&& [lindex $ChunkInfo 0] == "ChunkInfo"
		&&   [lindex $ChunkInfo 1] == $FileName_encoded
		&& [lindex $ChunkInfo 2] == $CurrentChunkNumber} { ;# proper response
		set ChunkChecksum [lindex $ChunkInfo 3]
	    } elseif {[eof $::s]} {
		return "Fail : Client is disconnected for ChunkInfo Downloading."
	    } else {
		puts "Fail : Fail to receive chunk information of $FileName."
		continue
	    }

	    # Receive the Chunk Data
	    set Chunk [open ShareBear.chk w]
	    fconfigure $Chunk -translation binary
	    puts $::s "QUERY_ChunkDownload $FileName_encoded $CurrentChunkNumber"
	    fconfigure $::s -buffering line -blocking 0 -translation binary -encoding binary
	    if {$CurrentChunkNumber < $TotalChunkNumber} {
		fcopy $::s $Chunk -size $::chunksize
	    } elseif {$CurrentChunkNumber == $TotalChunkNumber} {
		fcopy $::s $Chunk -size [expr $FileSize % $::chunksize ]
	    }
	    close $Chunk
	    fconfigure $::s -buffering line -blocking 0 -translation binary -encoding utf-8

	    # Check the chunk checksum
	    if {[crc::sum -file ShareBear.chk] == $ChunkChecksum} {
		# Copy received chunk to original file
		set Chunk [open ShareBear.chk r]
		fconfigure $Chunk -translation binary
		fcopy $Chunk $fp
		close $Chunk

		incr CurrentChunkNumber ;# Move to next chunk
	    } else {
		puts "Fail : Fail to receive chunk. Let's try again."
		continue
	    }
	}

	# One specific file upload is completed. File will be checked by checksum information of while loop.
	close $fp
    }
    file delete ShareBear.chk ;# remove temp chunk file at the end of process
    puts $::s "Info : File Download is completed." ;# Send message to server to complete the download operation.

    for {gets $::s line} {[string equal $line ""]} {gets $::s line} {after 100} ;# Read server's final response
    return $line ;# then return that.
}
