#!/usr/bin/tclsh
#
# Authors:
# 2008, SHY <abbagom@gmail.com>
# Authentication, File transfer over TLS, Revision Control.
# 2003, Reinhard Max
# SSL support added by Pat Thoyts <patthoyts@users.sourceforge.net>
#
# This file may be used and distributed under the same conditions as Tcl/Tk or GPL
#

# Set System encoding to use Korean well.
encoding system utf-8

# Log the information then print it on screen
proc log {String} {
    # Check log-file pointer is exist or not. If it set by other thread, wait it is closed.
    while {[info exists ::fp_log]} {
	after 100
    }
    set ::fp_log [open ShareBear_server.log a+]

    # Print message to log-file pointer, then unset variable on global.
    puts $::fp_log $String
    close $::fp_log
    unset ::fp_log

    # Print message to stdout or stderr.
    if {[lindex $String 0] == "Error" ||
	[lindex $String 0] == "Fail"} {
	puts stderr $String
    } else {
	puts stdout $String
    }
}

# Save the Array Data in DB format to DB File
proc SaveDB {DBName FileName} {
    # Naming file pointer
    set name_of_fp [format "%s_%s" "::fp" $FileName] ;# file pointer -> ::fp_FileName

    # Check file pointer is open or not by other process. If it is, wait for a while.
    while {[info exists $name_of_fp]} {
	after 100
    }

    # Make a backup, then open it.
    if {[file exist $FileName]} {
	file rename -force $FileName [format "%s.bak" $FileName]
    }
    set $name_of_fp [open $FileName w]

    # Print whole DB-contents to file, then unset file pointer.
    puts [set $name_of_fp] [list array set $DBName [array get $DBName]]
    close [set $name_of_fp]
    unset $name_of_fp

    # Close the proc
    return "INFO : Saving whole data to file, is completed."
}

# Read user configuration
# This configuration file parser is from Mike Tuxford on comp.lang.tcl in Dec. 2002. See the http://wiki.tcl.tk/2438 for mores.
proc parseConfig {configfile} {
    if {[file exists $configfile]} {
	set fd [open $configfile r]
    } else {
	log "Error : Failed to read configuration file. Check the file is exist or its permission."
	exit
    }
    while {![eof $fd]} {
	set data [gets $fd]
	# Skip the commented or null line
	if {[string index $data 0] == "#" || \
		[string index $data 0] == " " || \
		[string index $data 0] == ""} {
	    continue
	}
	# Set the "Config" array, using configuration file contents
	switch [lindex $data 0] {
	    port {
		global Config
		set Config(port) [lindex $data 1]
	    }
	    key_private {
		global Config
		set Config(key_private) [lindex $data 1]
	    }
	    key_public {
		global Config
		set Config(key_public) [lindex $data 1]
	    }
	    directory_file_save {
		global Config
		set Config(directory_file_save) [lindex $data 1]
	    }

	    default {
		global Config
		set Config([lindex $data 0]) [lrange $data 1 end]
	    }
	}
    }
    catch {close $fd}
    return
}
parseConfig ShareBear_server.cfg ;# Read configuration file : ShareBear_server.cfg

# Check the required operating environment is ready or not.
#     TLS(at least 1.5) extension, SSL private key, SSL public key should be ready
if {![catch {package require tls}] } {
    if {[file exists $Config(key_private)] && [file exists $Config(key_public)]} {
	::tls::init \
	    -certfile $Config(key_public) \
	    -keyfile  $Config(key_private) \
	    -ssl2 1 \
	    -ssl3 1 \
	    -tls1 0 \
	    -require 0 \
	    -request 0
    } else {
	log "Error : Failed to start. private and public key should be ready."
	log "        Try to make a key pair by following step on openssl(http://www.openssl.org)."
	log ""
	log "        openssl> genrsa -out <Private-key filename> 2048"
	log "        openssl> req -new -x509 -key <Private-key filename> -out <Public-key filename> -days <valid period of public key>"
	log ""
	exit
    }
} else {
    log "Error : Failed to start. tls extension should be installed."
    exit
}
#     tcllib is required.
if {[catch {package require otp}] ||
    [catch {package require md5crypt}] ||
    [catch {package require base64}] ||
    [catch {package require aes}] ||
    [catch {package require term::ansi::send}] ||
    [catch {package require sum}] ||
    [catch {package require uri::urn}]} {
    log "Error : Failed to start. tcllib is required to operate normally."
}
#    Load user account configuration. If it isn't ready, creat it.
if {[file exist user_db]} {
    if {[catch {source user_db}]} { ;# read current user account set
	log "Error : Failed to start. User account information is invalid."
	exit
    }
} else {
    # read admin password on STDIN
    puts -nonewline stdout "Set new admin's password: "
    ::term::ansi::send::sda_hidden ;#Hide the admin password on console
    gets stdin temp_admin_password
    ::term::ansi::send::sda_nohidden ;# Back to the normal display environment
    # make a admin token
    set md5_temp_admin_password [::md5crypt::md5crypt $temp_admin_password admin]
    set admin_token [::md5crypt::md5crypt $md5_temp_admin_password admin]
    # set initial uid
    set user_db(lastuid) 0
    # set admin information
    set uid [incr user_db(lastuid)]
    set user_db($uid,id) admin
    set user_db($uid,token) $admin_token
    set user_db($uid,permission_read) 1
    set user_db($uid,permission_write) 1
    set user_db($uid,permission_delete) 1
    set user_db($uid,e-mail) "" 
    set user_db($uid,etc) ""
    # set anonymous information
    set uid [incr user_db(lastuid)]
    set user_db($uid,id) anonymous
    set user_db($uid,token) {$1$anonymou$rl1SaWQRxBi4QIEEhBDPf.} ;#anonymous password->null
    set user_db($uid,permission_read) 0
    set user_db($uid,permission_write) 0
    set user_db($uid,permission_delete) 0
    set user_db($uid,e-mail) "" 
    set user_db($uid,etc) "anonymous user"
    # save the array data into file.
    SaveDB ::user_db user_db
    # unset the password variables, after all configuration is ended
    unset temp_admin_password
    unset md5_temp_admin_password
    unset admin_token
}
#    Load file server structure. If it isn't ready, creat it.
if {[file exist file_db]} {
    if {[catch {source file_db}]} { ;# read current file data information.
	log "Error : Failed to start. File storage database is not vaild."
	exit
    }
} else {
    # set initial fileid
    set file_db(LastFileID) 0

    # Make a pseudo structure
    #     set FileID [incr file_db(LastFileID)]
    #     set file_db($FileID,LastFileRev) 0
    #     set FileRev [incr file_db($FileID,LastFileRev)]

    #     set file_db($FileID,$FileRev,RegisteredTime) 20090217154133 ;#YYYYMMDDHHMMSS
    #     set file_db($FileID,$FileRev,UploaderID) {foobar}
    #     set file_db($FileID,$FileRev,UploaderIP) {127.0.0.1}
    #     set file_db($FileID,$FileRev,FileList) {foo.1 bar.2 foobar.3}
    #     set file_db($FileID,$FileRev,TagList) {linux GNU KLDP ShareBear Bear}
    #     set file_db($FileID,$FileRev,Description) {Pseudo description contents.}

    # save the array data into file.
    SaveDB ::file_db file_db
}

# Command list of file sharing protocol
set commands {
    bye
    showstate
    share_the_salt
    register_account
    authentication
    share_the_chunksize
    file_upload
    send_taglist
    send_filelist
    send_filename
    send_filedescription
    send_fileinformation
    file_download
    shutdown
}

# Server Commands : bye : Destroy the connection socket.
proc bye {} {
    upvar 1 state state
    after idle [list slaveServer::closeSocket $state(socket)]
    return "Good bye!"
}

# Server Commands : showstate : Return the status.
proc showstate {} {
    upvar 1 state state
    farray state
}

# Server Commands : share_the_salt : generate the salt for data encryption.
proc share_the_salt {} {
    upvar 1 state state
    # generate the salt by one-time-password
    set state(salt) [otp::otp-md5 -hex -count 99 -seed [clock clicks] $state(host)]
    return $state(salt)
}

# Server Commands : register_account
proc register_account {id encrypted_hashed_password token_client email etc} {
    upvar 2 user_db user_db
    upvar 1 state state

    # Check the integrity
    if {[string equal $state(salt) ""]} {
	return "Fail : No salt is ready for data communication"
    }
    set hashed_password [string trimright [aes::aes -dir decrypt -key "$state(salt)" -mode ecb [base64::decode $encrypted_hashed_password]] \0]
    set token_server [md5crypt::md5crypt $hashed_password $id]
    if {![string equal $token_client $token_server]} {
	return "Fail : Arrived password information is not valid."
    }

    # URL-Decoding id/email/etc after integity check/password generation
    set id [encoding convertfrom utf-8 [uri::urn::unquote $id]]
    set email [encoding convertfrom utf-8 [uri::urn::unquote $email]]
    set etc [encoding convertfrom utf-8 [uri::urn::unquote $etc]]

    # Append user data to account information DB.
    foreach i [array names user_db *,id] {
	if {[string equal $user_db($i) $id]} {
	    return "Fail : User ID is already registered."}}
    set uid [incr user_db(lastuid)]
    set user_db($uid,id) $id
    set user_db($uid,token) $token_client
    set user_db($uid,permission_read) 0
    set user_db($uid,permission_write) 0
    set user_db($uid,permission_delete) 0
    set user_db($uid,email) $email 
    set user_db($uid,etc) $etc

    # Append user data to account information file (user_db)
    SaveDB ::user_db user_db

    return "Pass : account - $id is created successfully."
}

# Server Commands : authentication
proc authentication {id encrypted_hashed_password token_client} {
    upvar 2 user_db user_db
    upvar 1 state state

    # Check the integrity
    if {[string equal $state(salt) ""]} {
	set state(user) anonymous
	set state(auth) 0
	return "Fail : No salt is ready for data communication"
    }
    set hashed_password [string trimright [aes::aes -dir decrypt -key "$state(salt)" -mode ecb [base64::decode $encrypted_hashed_password]] \0]
    set token_server [md5crypt::md5crypt $hashed_password $id]
    if {![string equal $token_client $token_server]} {
	set state(user) anonymous
	set state(auth) 0
	return "Fail : Arrived password information is not valid."
    }

    # URL-Decoding id/email/etc after integity check/password generation
    set id [encoding convertfrom utf-8 [uri::urn::unquote $id]]

    # Compare the registered user token and received token, to allow the login request.
    # User uid should be continuous, to avoid error for following routine.
    for {set i 1} {$i <= $user_db(lastuid)} {incr i} {
	if {[string equal $user_db($i,id) $id]} {
	    if {[string equal $user_db($i,token) $token_client]} {
		# Succeed for authentication
		set state(user) $id
		set state(auth) 1
		return "Pass : user $id is log-in successfully."
	    }
	}
    }

    # Finally, user cannot log-in successfully. 
    set state(user) anonymous
    set state(auth) 0
    return "Fail : There's no such an account, or your password is wrong."
}

# Server Commands : share_the_chunksize : chunk size for file transfer
proc share_the_chunksize {} {
    return $::Config(file_chunk_size)
}

# Server Commands : file_upload :  Ready to receive binary files.
####################################################################################################
# #     Client : <ask chunksize command> => Server						   #
# #     Client <= Server : <Chunk Size>								   #
# 												   #
# #     Client : file_upload <New/FID to revise> <file list> <tag list> <file description> => Server		   #
# #         For Each File									   #
# #         Client <= Server : QUERY_FileInfo <file name> FileSize FileChecksum			   #
# # 	Client : FileInfo <file name> <file size> <file checksum> => Server			   #
# #             For Each Chunk									   #
# #             Client <= Server : QUERY_ChunkInfo <file name> <chunk number> ChunkCheckSum	   #
# # 	    Client : ChunkInfo <file name> <chunk number> <chunk checksum>			   #
# #             Client <= Server : QUERY_ChunkUpload <file name> <chunk number>			   #
# # 	    Client : <Chunk> => Server								   #
# 												   #
# #     Uploaded file/chunk shall be confirmed by checksum data.				   #
####################################################################################################
proc file_upload {FIDtoRevise FileList TagList Description} { ;# Each item of filelist is url-encoded, to avoid white space problem of list.
    upvar 2 file_db file_db ;# file structure
    upvar 1 state state ;# current socket connection information
    upvar 1 socket socket ;# use the assigned socket for file transfer

    # FileName of FileList should be unique in their list. If it isnt' stop the process.
    if {[llength $FileList] != [llength [lsort -unique $FileList]]} {
	return "Fail : There're same filename in the list.."
    }

    # Exception Handling : FileList should be filename, not fully qualified filename
    foreach FileName $FileList {
	if { [file tail $FileName] != $FileName} {
	    return "Fail : FileList should be not fully qualified format."
	}
    }

    # Exception Handling : When TagList is not assigned(null), set it to _NotAssigned_
    if {[llength $TagList] == 0} {
	lappend TagList _NotAssigned_
    }

    # File Description should be url-encoded to be recognized as one element
    set Description [encoding convertfrom utf-8 [uri::urn::unquote $Description]]

    # Set the FileID, FileRevision and basic information, before receive the whole files.
    #     If Newly registere case, set fid new.
    #     If user send FID, check that is exist then increase the revision number of assigned FID.
    if {$FIDtoRevise == "New" ||
	$FIDtoRevise == "new" ||
	$FIDtoRevise == "NEW" } {
	set FileID [incr file_db(LastFileID)] ;# Add new FIleID at the end of last FileID.
	set file_db($FileID,LastFileRev) 0
	set FileRev [incr file_db($FileID,LastFileRev)]
    } elseif { $FIDtoRevise <= $file_db(LastFileID) } {
	set FileID $FIDtoRevise
	set FileRev [incr file_db($FileID,LastFileRev)]

	# Use the recent TagList for current TagList of current FileID, when TagList is filled by temporary value.
	if { [llength $TagList] == 1 && [lindex $TagList 0] == "_NotAssigned_" } {
	    set TagList $file_db($FileID,[expr $FileRev - 1 ],TagList)
	}

    } else {
	puts $socket "Error : Invalid FileID parameter to file upload. It should be new or existing FID number. "
    }
    set file_db($FileID,$FileRev,RegisteredTime) [clock format [clock second] -format %Y%m%d%H%M%S] ;#YYYYMMDDHHMMSS
    set file_db($FileID,$FileRev,UploaderID) $state(user)
    set file_db($FileID,$FileRev,UploaderIP) $state(host)
    set file_db($FileID,$FileRev,FileList) "___NotCompleted___"
    set file_db($FileID,$FileRev,TagList) $TagList
    set file_db($FileID,$FileRev,Description) $Description

    SaveDB ::file_db file_db

    # Make a directory to save the file
    set DirectoryName [file join $::Config(directory_file_save) [format %s_rev%s $FileID $FileRev]] ;# Directory name is <FileID>_rev<FileRevision>
    file mkdir $DirectoryName

    # Chunk for data transfer block
    set FullyQualifiedChunkFile [file join $::Config(directory_file_save) [format %s_rev%s.chk $FileID $FileRev]] ;# Chunk File will be located at configurated directory to save the file, in name of <FileID>_rev<FileRevision>.chk
    set Chunk [open $FullyQualifiedChunkFile w]
    close $Chunk

    # Read and store the data from client, one by one.
    foreach FileName $FileList {
	# FileName is url-encoded to avoid white space problem. Make a real filename to save its own name.
	set FileName_decoded [encoding convertfrom utf-8 [uri::urn::unquote $FileName]]
	lappend FileList_decoded $FileName_decoded

	# Set the full name of file, then make it as null to avoid null-file-checksum-trial-error
	set fully_qualified_filename [file join $::Config(directory_file_save) $DirectoryName $FileName_decoded]
	set fp [open $fully_qualified_filename w]
	close $fp

	# Acquire the file information : File Size, File Checksum
	puts $socket "QUERY_FileInfo $FileName FileSize FileChecksum"
	for {gets $socket FileInfo} {[string equal $FileInfo ""] && (![eof $socket])} {gets $socket FileInfo} {after 100} ;# Wait until some data is arrived.
	if {[llength $FileInfo] == 4 && [lindex $FileInfo 0] == "FileInfo"  && [lindex $FileInfo 1] == $FileName} { ;# proper response
	    set FileSize [lindex $FileInfo 2]
	    set FileChecksum [lindex $FileInfo 3]
	} elseif {[eof $socket]} {
	    return "Fail : Client is disconnected for FileInfo uploading."
	} else {
	    return "Fail : By insane response of file transfer protocol, failed to receive $FileName ( $FileName_decoded )."
	}

	# Try to receive the files until it is vaild.
	while { [crc::sum -file $fully_qualified_filename] != $FileChecksum } {
	    # Open the file Channel
	    set fp [open $fully_qualified_filename w]
	    fconfigure $fp -translation binary

	    # Calculate the total number of chunk, Current Chunk
	    set TotalChunkNumber [expr {int(ceil($FileSize/[format %f $::Config(file_chunk_size)]))}]
	    set CurrentChunkNumber 1

	    # In order, Receive & Check the chunk by checksum. If it isn't valid, request it once more.
	    while {$CurrentChunkNumber <= $TotalChunkNumber} {
		# Request & Receive the chunk Info
		puts $socket "QUERY_ChunkInfo $FileName $CurrentChunkNumber ChunkCheckSum"
		for {gets $socket ChunkInfo} {[string equal $ChunkInfo ""]  && (![eof $socket])} {gets $socket ChunkInfo} {after 100} ;# Wait until some data is arrived.
		if {[llength $ChunkInfo] == 4
		    && [lindex $ChunkInfo 0] == "ChunkInfo"
		    &&   [lindex $ChunkInfo 1] == $FileName
		    && [lindex $ChunkInfo 2] == $CurrentChunkNumber} { ;# proper response
		    set ChunkChecksum [lindex $ChunkInfo 3]
		} elseif {[eof $socket]} {
		    return "Fail : Client is disconnected for ChunkInfo uploading."
		} else {
		    puts $socket "Fail : Fail to receive chunk information of $FileName ( $FileName_decoded )."
		    continue
		}

		# Receive the Chunk Data
		set Chunk [open $FullyQualifiedChunkFile w]
		fconfigure $Chunk -translation binary
		puts $socket "QUERY_ChunkUpload $FileName $CurrentChunkNumber"
		fconfigure $socket -buffering line -blocking 0 -translation binary -encoding binary
		if {$CurrentChunkNumber < $TotalChunkNumber} {
		    fcopy $socket $Chunk -size $::Config(file_chunk_size)
		} elseif {$CurrentChunkNumber == $TotalChunkNumber} {
		    fcopy $socket $Chunk -size [expr $FileSize % $::Config(file_chunk_size) ]
		}
		close $Chunk
		fconfigure $socket -buffering line -blocking 0 -translation binary -encoding utf-8

		# Check the chunk checksum
		if {[crc::sum -file $FullyQualifiedChunkFile] == $ChunkChecksum} {
		    # Copy received chunk to original file
		    set Chunk [open $FullyQualifiedChunkFile r]
		    fconfigure $Chunk -translation binary
		    fcopy $Chunk $fp
		    close $Chunk

		    incr CurrentChunkNumber ;# Move to next chunk
		} else {
		    puts $socket "Fail : Fail to receive chunk of $FileName ( $FileName_decoded )."
		    continue
		}
	    }

	    # One specific file upload is completed. File will be checked by checksum information of while loop.
	    close $fp
	}
    }

    # Clear the chunk
    file delete $FullyQualifiedChunkFile

    # Make File-Information
    set FileInformation [open [file join $::Config(directory_file_save) [format %s_rev%s.txt $FileID $FileRev]] w] ;# File will be located at configurated directory to save the file, in name of <FileID>_rev<FileRevision>.txt
    puts $FileInformation "[clock format [clock scan $file_db($FileID,$FileRev,RegisteredTime) -format %Y%m%d%H%M%S]]" ;# beautified time description
    puts $FileInformation "User : $file_db($FileID,$FileRev,UploaderID)"
    puts $FileInformation "IP : $file_db($FileID,$FileRev,UploaderIP)"
    puts $FileInformation "FileList : $FileList_decoded"
    puts $FileInformation "TagList  : $file_db($FileID,$FileRev,TagList)"
    puts $FileInformation "Description : "
    puts $FileInformation $file_db($FileID,$FileRev,Description)
    close $FileInformation

    # Set the received file list, then save the current file_db to file.
    set file_db($FileID,$FileRev,FileList) $FileList

    SaveDB ::file_db file_db

    return "Info : File transfer is completed."
}

# Server Commands : send_taglist : Ask to send tags which were registered on fileserver.
proc send_taglist {} {
    # use the file_db 
    upvar 2 file_db file_db ;# file structure

    # Collect the all kinds of tag from file_db
    set TagList {}
    for {set FileID 1} {$FileID <= $file_db(LastFileID)} {incr FileID} {
	foreach Tag $file_db($FileID,$file_db($FileID,LastFileRev),TagList) {
	    lappend TagList $Tag
	}
    }

    # Post Processing : Make list unique, URL-Encoding to avoid white space problem.
    set TagList [lsort -unique $TagList]
    set TagList_encoded {}
    foreach Tag $TagList {
	lappend TagList_encoded [uri::urn::quote [encoding convertto utf-8 $Tag]]
    }

    # Return the TagLists
    if {[llength $TagList_encoded] == 0} {
	return "INFO : No Tag is available for now."
    } else {
	return $TagList_encoded
    }

}

# Server Commands : send_filelist : Ask to send FileID-RevisionID-FileList
proc send_filelist {TagList_client Selection_AndOr} {
    # use the file_db 
    upvar 2 file_db file_db ;# file structure

    # Exception Handling : When TagList is not assigned(null), set it to _NotAssigned_
    if {[llength $TagList_client] == 0} {
	lappend TagList_client _NotAssigned_
	set Selection_AndOr and
    }

    # set character to lower case
    set Selection_AndOr [string tolower $Selection_AndOr]

    # Find the FileID who has the Tag in assigned TagList.
    set FileIDList {}
    for {set FileID 1} {$FileID <= $file_db(LastFileID)} {incr FileID} { ;# For all the files
    	foreach Tag $file_db($FileID,$file_db($FileID,LastFileRev),TagList) { ;# For last revision
	    if {[lsearch -all -exact -inline $TagList_client $Tag] != ""} { ;# If any tags is same with assigned tag list
		lappend FileIDList $FileID
	    }
    	}
    }

    # Sort FileIDList in order. This could be handled differently by user selection:and/or
    if {$Selection_AndOr == "or"} { ;# If 'or', select file which has one of tag of client specified.
	set FileIDList [lsort -unique -integer $FileIDList] ;#neglect some same FileID.

    } elseif {$Selection_AndOr == "and"} { ;# If 'and', select file which has all tag element of client specified.
	set FileIDList_AndSelection {}
	for {set FileID 1} {$FileID <= $file_db(LastFileID)} {incr FileID} { ;# For all the files
	    if {[llength [lsearch -all -exact -inline $FileIDList $FileID]] == [llength $TagList_client]} { ;# Number of same FileID in FileIDList is same with Number of tag in user specifed taglist  => means FileID includes all tags which are defined by client.
		lappend FileIDList_AndSelection $FileID
	    }
	}
	set FileIDList $FileIDList_AndSelection

    } else { ;# argument should be and/or
	return "Error : use the AND/OR to get the filelist selection."
    }

    # Read the RevisionID/FileList using found FileID list
    if {[llength $FileIDList] == 0} {
	return "INFO : No File is available for assigned Tag List."
    } else {
	set FileID_FileRevision_FileList {}
	foreach FileID $FileIDList {
	    set FileRevision $file_db($FileID,LastFileRev)
	    set FileList [join $file_db($FileID,$FileRevision,FileList) ","] ;# Split the list by comma.
	    lappend FileID_FileRevision_FileList [format "FileID:%d FileRevision:%d FileList:%s" $FileID $FileRevision $FileList] ;# Make a list of formatted information : FileID-FileRevision-FileList
	}
    }

    # Return the TagLists
    return [join $FileID_FileRevision_FileList "\n"] ;# return each FileID-FileRevision-FileList information in different line.
}

# Server Commands : send_filename : Ask server to send filename of specified FileID, FileRevision.
proc send_filename {FileID FileRevision} {
    # use the file_db 
    upvar 2 file_db file_db ;# file structure

    # Exception Handling : Check File ID/Revision is exist or not.
    if {$FileID > $file_db(LastFileID)} {
	return "Error : Requested File ID is not exist."
    } elseif {$FileRevision > $file_db($FileID,LastFileRev)} {
	return "Error : Requested Revision is not exist."
    }

    # Return in separated list by white space format. Each Filename was url-encoded already.
    return [join $file_db($FileID,$FileRevision,FileList) " "]
}

# Server Commands : send_filedescription : Ask server to send assigned file description
proc send_filedescription {FileID FileRevision} {
    # use the file_db 
    upvar 2 file_db file_db ;# file structure

    # Exception Handling : Check File ID/Revision is exist or not.
    if {$FileID > $file_db(LastFileID)} {
	return "Error : Requested File ID is not exist."
    } elseif {$FileRevision > $file_db($FileID,LastFileRev)} {
	return "Error : Requested Revision is not exist."
    }

    # return url-encoded file description of assigned FileID/FileRevision
    return [uri::urn::quote [encoding convertto utf-8 $file_db($FileID,$FileRevision,Description)]]
}

# Server Commands : send_fileinformation : Request server to send overall information of registered file
proc send_fileinformation {FileID FileRevision} {
    # use the file_db 
    upvar 2 file_db file_db ;# file structure

    # Exception Handling : Check File ID/Revision is exist or not.
    if {$FileID > $file_db(LastFileID)} {
	return "Error : Requested File ID is not exist."
    } elseif {$FileRevision > $file_db($FileID,LastFileRev)} {
	return "Error : Requested Revision is not exist."
    }

    # Make fully qualified FileName
    set FileName_Information [file join $::Config(directory_file_save) [format %s_rev%s.txt $FileID $FileRevision]]

    # Exception Handling : Information file is not ready, when file transfer is not completed normally.
    if {![file exist $FileName_Information] || [file size $FileName_Information] == 0} {
	return "Fail : Requested file information is not ready."
    }

    # Open the information file of FileID/FileRevision
    set fp [open $FileName_Information r]
    set FileInformation [read $fp]
    close $fp

    # URL encoding to send informations in one line.
    set FileInformation_encoded [uri::urn::quote [encoding convertto utf-8 $FileInformation]]

    # return 
    return $FileInformation_encoded
}

# Server Commands : file_download :  Ready to download file which was requested by client
##############################################################################################################
# #     Client : <ask chunksize command> => Server						   	     #
# #     Client <= Server : <Chunk Size>									     #
# 												   	     #
# #     Client : file_download <FileID> <FileRevision> <FileName> => Server				     #
# #         Client <= Server : FileInfo <FileName> <FileSize> <FileChecksum>				     #
# # 	    For Each Chunk										     #
# #             Client : QUERY_ChunkInfo <FileName> <ChunkNumber> ChunkCheckSum => Server		     #
# # 	    Client <= Server : ChunkInfo <FileName> <ChunkNumber> <ChunkCheckSum>			     #
# #             Client : QUERY_ChunkDownload <FileName> <ChunkNumber> => Server				     #
# # 	    Client <= Server : <Chunk>									     #
# 													     #
# #     File downloading function handles just one file at once.					     #
##############################################################################################################
proc file_download {FileID FileRevision FileName} {
    upvar 1 socket socket ;# use the assigned socket for file transfer

    # Make fully qualified file name, then check it is exist or not.
    set FileName_decoded [encoding convertfrom utf-8 [uri::urn::unquote $FileName]]
    set DirectoryName [file join $::Config(directory_file_save) [format %s_rev%s $FileID $FileRevision]] ;# Directory name is <FileID>_rev<FileRevision> @ configured storage directory
    set FullyQualifiedFileName [file join $DirectoryName $FileName_decoded]
    if {![file exist $FullyQualifiedFileName]} {
	return "Error : Not existed file is requested to download."
    }

    # Make FileSize/FileChecksum info, then send it to client.
    set FileSize [file size $FullyQualifiedFileName]
    set FileChecksum [crc::sum -file $FullyQualifiedFileName]
    puts $socket "FileInfo $FileName $FileSize $FileChecksum"

    # Until file transfer is completed or socket is closed, make a chunk/chunk checksum then send it to client by client request.
    for {gets $socket Rx_buffer} {[string equal $Rx_buffer ""]} {gets $socket Rx_buffer} {after 100}
    while {$Rx_buffer != "Info : File Download is completed." } {
	if {[eof $socket]} { ;# Complete the filedownloading when socket is disconnected.
	    return "Fail : Client is disconnected while file downloading."
	} elseif {[lindex $Rx_buffer 0] == "QUERY_ChunkInfo"} {
	    set ChunkNumber [lindex $Rx_buffer 2]
	    set ChunkFileName [GetChunkToDownload $FullyQualifiedFileName $::Config(file_chunk_size) $ChunkNumber $socket]
	    set ChunkChecksum [crc::sum -file $ChunkFileName]
	    file delete $ChunkFileName

	    puts $socket "ChunkInfo [lindex $Rx_buffer 1] [lindex $Rx_buffer 2] $ChunkChecksum"

	} elseif {[lindex $Rx_buffer 0] == "QUERY_ChunkDownload"} {
	    set ChunkNumber [lindex $Rx_buffer 2]
	    set ChunkFileName [GetChunkToDownload $FullyQualifiedFileName $::Config(file_chunk_size) $ChunkNumber $socket]

	    # Open chunk file, and copy that to socket in binary encoding.
	    set Chunk [open $ChunkFileName r]
	    fconfigure $Chunk -translation binary
	    fconfigure $socket -buffering line -blocking 0 -translation binary -encoding binary
	    fcopy $Chunk $socket
	    close $Chunk
	    fconfigure $socket -buffering line -blocking 0 -translation binary -encoding utf-8
	    file delete $ChunkFileName
	}

	# Read Command from Server, for next process
	for {gets $socket Rx_buffer} {[string equal $Rx_buffer ""]} {gets $socket Rx_buffer} {after 100}
    }
    return "Info : Download is completed."
}
proc GetChunkToDownload {FileName ChunkSize ChunkNumber ClientSocketName} {
    # Open Filestream and configure that.
    set fp [open $FileName r] ;# Fully qualified filename
    set ChunkFileName [format "%s_%s" ShareBear.chk $ClientSocketName]  ;# ChunkFileName should be unique, by socket number.
    set Chunk [open $ChunkFileName w]

    fconfigure $fp -encoding binary -translation binary
    fconfigure $Chunk -encoding binary -translation binary
    
    seek $fp [expr (($ChunkNumber - 1) * $ChunkSize)] start

    # Calculate the total number of chunk
    set FileSize [file size $FileName]
    set TotalChunkNumber [expr {int(ceil($FileSize/[format %f $ChunkSize]))}]

    # Copy the Chunk from File
    if {$ChunkNumber < $TotalChunkNumber} {
	fcopy $fp $Chunk -size $ChunkSize
    } elseif {$ChunkNumber == $TotalChunkNumber} {
	fcopy $fp $Chunk -size [expr $FileSize % $ChunkSize ]
    } else {
	return "Error : invalid chunk number was given."
    }

    # Close Filestream
    close $fp
    close $Chunk

    # Return the Unique ChunkFile Name
    return $ChunkFileName
}

# Server Commands : shutdown : Terminate the server program.
proc shutdown {} {
    upvar 1 socket socket ;# use the assigned socket for file transfer
    set termination_message "INFO : ShutDown. : [clock format [clock second] -format %Y%m%d%H%M]"
    puts $socket $termination_message
    flush $socket

    log $termination_message

    after 3000
    exit
}

# showstate related procedures.
proc max {a b} {expr {$a > $b ? $a : $b}}
proc farray {array {separator =} {pattern *}} {
    upvar $array a
    set names [lsort [array names a $pattern]]
    set max 0
    foreach name $names {
        set max [max $max [string length $name]]
    }
    set result [list]
    foreach name $names {
        lappend result [format " %-*s %s %s" $max $name $separator $a($name)]
    }
    return [join $result "\n"]
}

namespace eval slaveServer {
    # procs that start with a lowercase letter are public
    namespace export {[a-z]*}
    variable serversocket
}

proc slaveServer::closeSocket {socket} {
    variable $socket
    upvar 0 $socket state
    log "Closing $socket [clock format [clock seconds]]"
    catch {close $socket}
    unset state
}

# This gets called whenever a client connects
proc slaveServer::Server {socket host port} {
    variable $socket
    upvar 0 $socket state
    # just to be sure ...
    array unset state
    set state(socket) $socket
    set state(host) $host
    set state(port) $port
    set state(user) "anonymous"
    set state(salt) ""
    set state(auth) 0
    log "New Connection: $socket $host $port [clock format [clock seconds]]"
    fconfigure $socket -buffering line -blocking 0 -translation binary -encoding utf-8
    fileevent $socket readable [list [namespace code Handler] $socket]
    puts $socket "Welcome to this little file sharing server!"
    puts $socket ""
}

# This gets called whenever a client sends a new line
# of data or disconnects
proc slaveServer::Handler {socket} {
    variable $socket
    upvar 0 $socket state

    # Do we have a disconnect?
    if {[eof $socket]} {
        closeSocket $socket
        return
    }

    # Does reading the socket give us an error?
    if {[catch {gets $socket line} ret] == -1} {
        log "Closing $socket"
        closeSocket $socket
        return
    }
    # Did we really get a whole line?
    if {$ret == -1} return

    # ... and is it not empty? ...
    set line [string trim $line]
    if {$line == ""} return

    ## ... and not an SSL request? ...
    #if {[string index $line 0] == "\200"} {
    #    log "SSL request - closing connection"
    #    closeSocket $socket
    #    return
    #}

    # OK, so log it ...
    log "$socket [clock format [clock second] -format %Y%m%d%H%M] > $line"

    # ... evaluate it, ...
    set line [regsub -all {\$} $line {\\$}]
    set line [regsub -all {\[} $line {\\[}]
	      set line [regsub -all {\]} $line {\\]}]
    if {[catch {slave eval "$line"} ret]} {
        set ret "ERROR: $ret"
    }
    # ... log the result ...
    log [regsub -all -line ^ $ret "$socket [clock format [clock second] -format %Y%m%d%H%M] < "]

    # ... and send it back to the client.
    if {[catch {puts $socket $ret}]} {
        closeSocket $socket
    }
}

proc slaveServer::init {port commands} {
    variable serversockets
    # Start Message
    log "File Sharing Server is started on [clock format [clock second] -format %Y%m%d%H%M]"
    # (re-)create a safe slave interpreter
    catch {interp delete slave}
    interp create -safe slave
    # remove all predefined commands from the slave
    foreach command [slave eval info commands] {
        slave hide $command
    }
    # link the commands for the protocol into the slave
    foreach command $commands {
        log "Initializing Command : $command"
        interp alias slave $command {} $command
    }
    #(re-)create the server socket
    if {[info exists serversockets]} {
        foreach sock $serversockets {
            catch {close $sock}
        }
        unset serversockets
    }
    log "Opening socket: $port ::tls::socket"
    lappend serversockets [::tls::socket -server [namespace code Server] $port]
    log "Opening socket: Succeed."
}

slaveServer::init $Config(port) $commands
if {![info exists forever]} {
    set forever 1
    vwait forever
}